@@ -1,6 +1,9 @@
-import os
+from collections import namedtuple
+from typing import Optional, Any, Generator, List
 import json
+import os
 import requests
+import time
 
 from kbase_workspace_client.contigset_to_fasta import contigset_to_fasta
 from kbase_workspace_client.exceptions import (
@@ -12,8 +15,34 @@ from kbase_workspace_client.exceptions import (
     InvalidGenome,
 )
 
+# Named tuples for object info and workspace info
+ObjInfo = namedtuple('ObjInfo', [
+  "objid",
+  "name",
+  "type",
+  "save_date",
+  "version",
+  "saved_by",
+  "wsid",
+  "workspace",
+  "chsum",
+  "size",
+  "meta",
+])
+WSInfo = namedtuple('WSInfo', [
+  "id",
+  "workspace",
+  "owner",
+  "moddate",
+  "max_objid",
+  "user_permission",
+  "globalread",
+  "lockstat",
+  "metadata",
+])
 
-def _post_req(payload, url, token, file_path=None):
+
+def _post_req(payload: dict, url: str, token: Optional[str], file_path: str = None) -> Any:
     """Make a post request to the workspace server and process the response."""
     headers = {'Authorization': token}
     with requests.post(url, data=json.dumps(payload), headers=headers, stream=True) as resp:
@@ -34,7 +63,7 @@ def _post_req(payload, url, token, file_path=None):
             return resp_json['result'][0]
 
 
-def _validate_file_for_writing(dest_path):
+def _validate_file_for_writing(dest_path: str) -> None:
     """Validate that a path points to a non-existent file in a writable directory."""
     if os.path.isfile(dest_path):
         raise IOError(f"File path already exists: {dest_path}")
@@ -47,38 +76,49 @@ def _validate_file_for_writing(dest_path):
 
 class WorkspaceClient:
 
-    def __init__(self, url, token=None):
+    def __init__(self, url: str, token: str = None):
         """
         Instantiate the workspace client.
         Args:
-            url - URL of the workspace service with the root path
-            token - User or service authentication token from KBase. Optional.
+            url: URL of the workspace service with the root path
+            token: User or service authentication token from KBase. Optional.
         """
         self._url = url.strip('/')
         self._ws_url = url + '/ws'
         self._token = token
 
-    def req(self, method, params):
+    def req(self, method: str, params: dict) -> Any:
         """
         Make a normal request to the workspace.
         Args:
-            method - workspace method name (must be a funcdef in the KIDL spec)
-            params - parameters as python dicts, lists, and values
-        Returns python data (dicts/lists) of response data from the workspace.
-        Raises a WorkspaceResponseError on an unsuccessful request.
+            method: workspace method name (must be a funcdef in the KIDL spec)
+            params: parameters as python dicts, lists, and values
+        Returns:
+            python data (dicts/lists) of response data from the workspace.
+        Raises:
+            WorkspaceResponseError on an unsuccessful request.
         """
-        payload = {'version': '1.1', 'method': method, 'params': [params]}
+        _id = int(time.time() * 1000)
+        payload = {'version': '1.1', 'id': _id, 'method': method, 'params': [params]}
         return _post_req(payload, self._ws_url, self._token)
 
-    def generate_obj_infos(self, wsid, minid=1, maxid=None, latest=True, admin=False):
+    def generate_obj_infos(
+            self,
+            wsid: int,
+            minid: int = 1,
+            maxid: Optional[int] = None,
+            latest: bool = True,
+            admin: bool = False) -> Generator[list, None, None]:
         """
         Generator, yielding all object IDs + version IDs in a workspace.
         This handles the 10k pagination and will generate *all* ids.
         Args:
-            wsid - int - workspace id (int)
-            latest - bool - default True - Generate only the latest version of each obj.
-            admin - bool - default False - Make the "list_objects" request as an admin request.
-        Yields object ids
+            wsid: workspace ID
+            latest: Generate only the latest version of each obj, or generate
+                all versions of all objects.
+            admin: Make the "list_objects" request as a Workspace administrator
+        Yields:
+            Object info tuples (as python lists)
         """
         params = {"ids": [wsid]}  # type: dict
         if maxid:
@@ -97,14 +137,16 @@ class WorkspaceClient:
             for obj_info in part:
                 yield obj_info
 
-    def admin_req(self, method, params):
+    def admin_req(self, method: str, params: dict) -> Any:
         """
         Make a special workspace admin command.
         Args:
-            method - workspace method name (must be a funcdef in the KIDL spec)
+            method: workspace method name (must be a funcdef in the KIDL spec)
             params - parameters as python dicts, lists, and values
-        Returns python data (dicts/lists) of response data from the workspace.
-        Raises a WorkspaceResponseError on an unsuccessful request.
+        Returns:
+            python data (dicts/lists) of response data from the workspace.
+        Raises:
+            WorkspaceResponseError on an unsuccessful request.
         """
         payload = {
             'version': '1.1',
@@ -113,29 +155,33 @@ class WorkspaceClient:
         }
         return _post_req(payload, self._ws_url, self._token)
 
-    def req_download(self, method, params, dest_path):
+    def req_download(self, method: str, params: dict, dest_path: str) -> None:
         """
         Make a workspace request and download the response to a file (streaming)
         Args:
             method - workspace method name (must be a funcdef in the KIDL spec)
             params - parameters as python dicts, lists, and values
             dest_path - filepath where you would like to write out results
-        Returns None when the request is complete and the file is written.
-        Raises a WorkspaceResponseError on an unsuccessful request.
+        Returns:
+            None when the request is complete and the file is written.
+        Raises:
+            WorkspaceResponseError on an unsuccessful request.
         """
         _validate_file_for_writing(dest_path)
         payload = {'version': '1.1', 'method': method, 'params': [params]}
         _post_req(payload, self._ws_url, self._token, dest_path)
 
-    def admin_req_download(self, method, params, dest_path):
+    def admin_req_download(self, method: str, params: dict, dest_path: str) -> None:
         """
         Make an admin command and download the response to a file (streaming)
         Args:
             method - workspace method name (must be a funcdef in the KIDL spec)
             params - parameters as python dicts, lists, and values
             dest_path - filepath where you would like to write out results
-        Returns None when the request is complete and the file is written.
-        Raises a WorkspaceResponseError on an unsuccessful request.
+        Returns:
+            None when the request is complete and the file is written.
+        Raises:
+            WorkspaceResponseError on an unsuccessful request.
         """
         _validate_file_for_writing(dest_path)
         payload = {
@@ -143,14 +189,15 @@ class WorkspaceClient:
             'method': 'Workspace.administer',
             'params': [{'command': method, 'params': params}]
         }
-        return _post_req(payload, self._ws_url, self._token, dest_path)
+        _post_req(payload, self._ws_url, self._token, dest_path)
 
-    def handle_to_shock(self, handle):
+    def handle_to_shock(self, handle: str) -> str:
         """
         Convert a handle ID to a shock ID
         Args:
-            handle
-        Returns the shock ID as a string
+            handle: handle service ID
+        Returns:
+            The shock node ID
         """
         headers = {'Content-Type': 'application/json'}
         if self._token:
@@ -169,14 +216,16 @@ class WorkspaceClient:
             raise RuntimeError(f"Error from handle_service: {resp.text}")
         return resp.json()['result'][0][0]['id']
 
-    def download_shock_file(self, shock_id, dest_path):
+    def download_shock_file(self, shock_id: str, dest_path: str) -> None:
         """
         Download a file from shock.
         Args:
             shock_id
             dest_path
-        Returns None when the file finishes downloading
-        Raises UnauthorizedShockDownload or MissingShockFile on failure
+        Returns:
+            None when the file finishes downloading
+        Raises:
+            UnauthorizedShockDownload or MissingShockFile on failure
         """
         _validate_file_for_writing(dest_path)
         headers = {'Authorization': ('OAuth ' + self._token) if self._token else None}
@@ -199,13 +248,15 @@ class WorkspaceClient:
                 for block in resp.iter_content(1024):
                     fwrite.write(block)
 
-    def download_assembly_fasta(self, ref, save_dir, admin=False):
+    def download_assembly_fasta(self, ref: str, save_dir: str, admin: bool = False) -> str:
         """
         Download an Assembly object as fasta.
-        Keyword arguments:
-          ref is a workspace reference ID in the form 'workspace_id/object_id/version'
-          save_dir is the path of a directory in which to save the fasta file
-        Returns an absolute path of the downloaded fasta file.
+        Args:
+            ref: a workspace reference ID in the form 'workspace_id/object_id/version'
+            save_dir: the path of a directory in which to save the fasta file
+            admin: whether to make the request as a Workspace administrator
+        Returns:
+            an absolute path of the downloaded fasta file.
         """
         ws_obj = _download_obj(self, ref, admin=admin)
         valid_types = ['KBaseGenomeAnnotations.Assembly', 'KBaseGenomes.ContigSet']
@@ -228,13 +279,9 @@ class WorkspaceClient:
             self.download_shock_file(shock_id, output_path)
         return output_path
 
-    def download_reads_fastq(self, ref, save_dir, admin=False):
+    def download_reads_fastq(self, ref: str, save_dir: str, admin: bool = False) -> List[str]:
         """
         Download genome reads data as fastq.
-        Keyword arguments:
-          ref is a workspace reference ID in the form 'workspace_id/object_id/version'
-          save_dir is the path of a directory in which to save the fasta file
-        Returns a list of paths of the downloaded fastq files.
 
         If the reads are paired-end and non-interleaved, you will get two files, one for the forward
         (left) reads and one for the reverse (right) reads. Otherwise, you will get one file.
@@ -244,6 +291,12 @@ class WorkspaceClient:
         - Paired ends and non-interleaved get the file ending of '.paired.fwd.fastq' and
             '.paired.rev.fastq'
         - Single ends get the file ending of '.single.fastq'
+
+        Keyword arguments:
+            ref: a workspace reference ID in the form 'workspace_id/object_id/version'
+            save_dir: the path of a directory in which to save the fasta file
+        Returns:
+            a list of paths of the downloaded fastq files.
         """
         # Fetch the workspace object and check its type
         ws_obj = _download_obj(self, ref, admin=admin)
@@ -281,12 +334,13 @@ class WorkspaceClient:
         output_paths = map(lambda pair: pair[1], to_download)
         return list(output_paths)
 
-    def get_assembly_from_genome(self, ref, admin=False):
+    def get_assembly_from_genome(self, ref: str, admin: bool = False) -> str:
         """
         Given a Genome object, fetch the reference to its Assembly object on the workspace.
-        Arguments:
-          ref is a workspace reference ID in the form 'workspace_id/object_id/version'
-        Returns a workspace reference to an assembly object
+        Args:
+            ref is a workspace reference ID in the form 'workspace_id/object_id/version'
+        Returns:
+            workspace reference to an assembly object
         """
         # Fetch the workspace object and check its type
         ws_obj = _download_obj(self, ref, admin=admin)
@@ -300,8 +354,32 @@ class WorkspaceClient:
         ref_path = ref + ';' + assembly_ref
         return ref_path
 
+    def find_narrative(self, wsid: int, admin: bool = False) -> Optional[ObjInfo]:
+        """
+        Fetch the narrative object out of a workspace.
+        Args:
+            wsid: workspace ID
+            admin: whether to make the request as a Workspace administrator
+        Returns:
+            None if no narrative present, or an ObjInfo for the narrative object.
+        """
+        req = self.admin_req if admin else self.req
+        ws_meth = "getWorkspaceInfo" if admin else "get_workspace_info"
+        ws_info_raw = req(ws_meth, {"id": wsid})
+        ws_info = WSInfo(*ws_info_raw)
+        # Fetch the narrative ID from the workspace metadata
+        try:
+            narr_obj_id = int(ws_info.metadata.get('narrative'))
+        except TypeError:
+            # No narrative ID accessible from the workspace info
+            return None
+        ref = f"{wsid}/{narr_obj_id}"
+        obj_meth = "getObjectInfo" if admin else "get_object_info3"
+        narr_info_raw = req(obj_meth, {"objects": [{"ref": ref}]})["infos"][0]
+        return ObjInfo(*narr_info_raw)
 
-def _download_obj(client, ref, data=True, admin=False):
+
+def _download_obj(client, ref: str, data: bool = True, admin: bool = False) -> dict:
     """Download an object with get_objects2."""
     params = {'objects': [{'ref': ref}]}  # type: dict
     if not data:
@@ -313,14 +391,14 @@ def _download_obj(client, ref, data=True, admin=False):
     return ws_obj['data'][0]
 
 
-def _validate_obj_type(ws_obj, types):
+def _validate_obj_type(ws_obj: dict, types: List[str]) -> None:
     """
     Given a workspace object, validate that its types match any of the strings in `types`.
     Args:
-      ws_obj is a workspace reference in the form of 'workspace_id/object_id/version'
-      types is a list of string type names to match against
-    raises an InvalidWSType error
-    returns None
+        ws_obj: a workspace reference in the form of 'workspace_id/object_id/version'
+        types: a list of string type names to match against
+    Raises:
+        InvalidWSType error if given object not in type list
     """
     ws_type = ws_obj['info'][2]
     if all(t not in ws_type for t in types):
